"""
Lump 15 - Worldlights
=====================

This lump contains an array of :any:`dworldlight_tV0` or :any:`dworldlight_tV1` (Version detection is automatic).
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402
from valvebsp.structs.flags import dworldlight_flags32  # NOQA: #402

emittype_t = Enum(
    Int32ul,
    emit_surface=0,
    emit_point=1,
    emit_spotlight=2,
    emit_skylight=3,
    emit_quakelight=4,
    emit_skyambient=5)

dworldlight_tV0 = Struct(
    'origin' / Vector,
    'intensity' / Vector,
    'normal' / Vector,
    'cluster' / Int32sl * "index into :ref:`lump 23<lump_23>`",
    'type' / emittype_t,
    'style' / Int32sl,
    'stopdot' / Float32l,
    'stopdot2' / Float32l,
    'exponent' / Float32l,
    'radius' / Float32l,
    'constantAttn' / Float32l,
    'linearAttn' / Float32l,
    'quadraticAttn' / Float32l,
    'flags' / dworldlight_flags32,
    'texinfo' / Int32sl * "index into :ref:`lump 6<lump_6>`",
    'owner' / Int32sl * "index into :ref:`lump 0<lump_0>`"
)

dworldlight_tV1 = Struct(
    'origin' / Vector,
    'intensity' / Vector,
    'normal' / Vector,
    'shadow_cast_offset' / Vector,
    'cluster' / Int32sl * "index into :ref:`lump 23<lump_23>`",
    'type' / emittype_t,
    'style' / Int32sl,
    'stopdot' / Float32l,
    'stopdot2' / Float32l,
    'exponent' / Float32l,
    'radius' / Float32l,
    'constantAttn' / Float32l,
    'linearAttn' / Float32l,
    'quadraticAttn' / Float32l,
    'flags' / dworldlight_flags32,
    'texinfo' / Int32sl * "index into :ref:`lump 6<lump_6>`",
    'owner' / Int32sl * "index into :ref:`lump 0<lump_0>`"
)


@lump_array
@lump_version([0, 1])
def lump_15(header, profile=None):
    if header.version == 0:
        return dworldlight_tV0
    elif header.version == 1:
        return dworldlight_tV1



"""
Lump 41 - Clip Portal Verts
===========================

This lump contains an array of verts represented as :any:`Vector`.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from valvebsp.structs.common import *  # NOQA: #402


@lump_array
@lump_version(0)
def lump_41(header, profile=None):
    return Vector



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()


class LumpUnsupportedError(Exception):
    pass


class LumpVersionUnsupportedError(Exception):
    pass


class LumpCompressedError(Exception):
    pass



"""
Lump 19 - Brush Sides
=====================

This lump contains an array of :any:`dbrushside_t`.

Note: Portal 2's implementation of this lump differs, a :ref:`profile<profiles>` must be specified.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402

dbrushside_t_P2 = Struct(
    'planeNum' / Int16ul,
    'texInfo' / Int16sl * "index into :ref:`lump 6<lump_6>`",
    'dispInfo' / Int16sl * "index into :ref:`lump 26<lump_26>`",
    'bevel' / Int8sl,
    'thin' / Int8sl
)

dbrushside_t = Struct(
    'planeNum' / Int16ul,
    'texInfo' / Int16sl * "index into :ref:`lump 6<lump_6>`",
    'dispInfo' / Int16sl * "index into :ref:`lump 26<lump_26>`",
    'bevel' / Int16sl
)


@lump_array
@lump_version(0)
def lump_19(header, profile=None):
    if profile in [ALIENSWARM, PORTAL2]:
        return dbrushside_t_P2
    else:
        return dbrushside_t



"""
Lump 54 - Worldlights HDR
=========================

This lump contains an array of :any:`dworldlight_tV0` or :any:`dworldlight_tV1` (Version detection is automatic).
It is the HDR version of :ref:`lump 15<lump_15>`."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from valvebsp.structs.common import *  # NOQA #402
from valvebsp.structs.lump_15 import dworldlight_tV0, dworldlight_tV1  # NOQA #402


@lump_array
@lump_version([0, 1])
def lump_54(header, profile=None):
    if header.version == 0:
        return dworldlight_tV0
    elif header.version == 1:
        return dworldlight_tV1



"""
Lump 1 - Planes
===============

This lump contains an array of :any:`dplane_t`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import Struct, Int32sl, Float32l  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402


dplane_t = Struct(
    'normal' / Vector,
    'dist' / Float32l,
    'type' / Int32sl
)


@lump_array
@lump_version(0)
def lump_1(header, profile=None):
    return dplane_t



"""
Lump 35 - Game Lump
===================

This lump contains a single :any:`dgamelumpheader_t`. This is only the gamelump headers and not the gamelump data. for gamelump data, :ref:`see gamelumps<gamelumps>`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA #402
from valvebsp.structs.flags import dgamelump_flags16  # NOQA: #402


dgamelump_t = Struct(
    'id' / PaddedString(4, "ascii"),
    'flags' / dgamelump_flags16,
    'version' / Int16ul,
    'fileofs' / Int32sl,
    'filelen' / Int32sl,
)

dgamelumpheader_t = Struct(
    'lumpCount' / Int32sl,
    'gamelump' / dgamelump_t[this.lumpCount],
)


@lump_struct
@lump_version(0)
def lump_35(header, profile=None):
    return dgamelumpheader_t



"""
Lump 0 - Entities
=================

This lump is simply one long string containing all entities in a format similar to vmf.

ValveBsp will parse this string into a list of entities, represented as lists of properties.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
import pyparsing as pp   # NOQA: E402
from valvebsp.structs.common import *  # NOQA: #402


entity_property_encoding = pp.Group(pp.QuotedString('"') * 2)
entity_encoding = pp.nestedExpr('{', '}', entity_property_encoding, None)
lump_0_encoding = pp.ZeroOrMore(entity_encoding)


class EntityEncoder(Adapter):

    def _decode(self, obj, context, path):
        return lump_0_encoding.parseString(obj).asList()

    def _encode(self, obj, context, path):
        estr = ''
        for ent in obj:
            estr += '{\n'
            for entp in ent:
                estr += '"' + entp[0] + '" "' + entp[1] + '"\n'
            estr += '}\n'
        estr += '\0'
        return estr


@lump_version(0)
def lump_0(header, profile=None):
    return EntityEncoder(GreedyString("ascii"))



"""
Lump 12 - Edges
===============

This lump contains an array of :any:`dedge_t`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402

dedge_t = Struct(
    'v' / Int16ul[2] * "index into :ref:`lump 3<lump_3>`"
)


@lump_array
@lump_version(0)
def lump_12(header, profile=None):
    return dedge_t



"""
Lump 10 - Leafs
===============

This lump contains an array of :any:`dleaf_tV1` or :any:`dleaf_tV1` (Version detection is automatic).
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402
from valvebsp.structs.flags import contents_flags32, dleaf_flags7  # NOQA: #402

dleaf_tV0 = Aligned(4, Bitwise(Struct(
    'contents' / Bytewise(contents_flags32),
    'cluster' / Bytewise(Int16sl),

    'area' / BitsInteger(9),
    'flags' / dleaf_flags7,

    'mins' / Bytewise(Int16sl[3]),
    'maxs' / Bytewise(Int16sl[3]),

    'firstLeafFace' / Bytewise(Int16ul) * "index into :ref:`lump 16<lump_16>`",
    'numLeafFaces' / Bytewise(Int16ul),

    'firstLeafBrush' / Bytewise(Int16ul) * "index into :ref:`lump 17<lump_17>`",
    'numLeafBrushes' / Bytewise(Int16ul),
    'leafWaterDataID' / Bytewise(Int16sl) *
    "index into :ref:`lump 36<lump_36>`",

    'ambientLighting' / Bytewise(CompressedLightCube)
)))

dleaf_tV1 = Aligned(4, Bitwise(Struct(
    'contents' / Bytewise(contents_flags32),
    'cluster' / Bytewise(Int16sl),

    'area' / BitsInteger(9),
    'flags' / dleaf_flags7,

    'mins' / Bytewise(Int16sl[3]),
    'maxs' / Bytewise(Int16sl[3]),


    'firstLeafFace' / Bytewise(Int16ul) * "index into :ref:`lump 16<lump_16>`",
    'numLeafFaces' / Bytewise(Int16ul),

    'firstLeafBrush' / Bytewise(Int16ul) * "index into :ref:`lump 17<lump_17>`",
    'numLeafBrushes' / Bytewise(Int16ul),
    'leafWaterDataID' / Bytewise(Int16sl) * "index into :ref:`lump 36<lump_36>`"
)))


@lump_array
@lump_version([0, 1])
def lump_10(header, profile=None):
    if header.version == 0:
        return dleaf_tV0
    elif header.version == 1:
        return dleaf_tV1




"""
Lump 2 - Texture Data
=====================

This lump contains an array of :any:`dtexdata_t`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402

dtexdata_t = Struct(
    'reflectivity' / Vector * "RGB reflectivity",
    'nameStringTableID' / Int32sl * "index into :ref:`lump 44<lump_44>`",
    'width' / Int32sl,
    'height' / Int32sl,
    'viewWidth' / Int32sl,
    'viewHeight' / Int32sl
)


@lump_array
@lump_version(0)
def lump_2(header, profile=None):
    return dtexdata_t



"""
Lump 18 - Brushes
=================

This lump contains an array of :any:`dbrush_t`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402
from valvebsp.structs.flags import contents_flags32  # NOQA: #402

dbrush_t = Struct(
    'firstSide' / Int32sl * "index into :ref:`lump 19<lump_19>`",
    'numSides' / Int32sl,
    'contents' / contents_flags32,
)


@lump_array
@lump_version(0)
def lump_18(header, profile=None):
    return dbrush_t



"""
Lump prpd - Prop Detail
=======================

This lump is defined as a single :any:`DetailPropLump_t`
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402

DetailPropOrientation_t = Enum(
    Int8ul,
    DETAIL_PROP_ORIENT_NORMAL=0,
    DETAIL_PROP_ORIENT_SCREEN_ALIGNED=1,
    DETAIL_PROP_ORIENT_SCREEN_ALIGNED_VERTICAL=2,
)

DetailPropType_t = Enum(
    Int8ul,
    DETAIL_PROP_TYPE_MODEL=0,
    DETAIL_PROP_TYPE_SPRITE=1,
    DETAIL_PROP_TYPE_SHAPE_CROSS=2,
    DETAIL_PROP_TYPE_SHAPE_TRI=3,
)

DetailSpriteLump_t = Struct(
    'UL' / Vector2D,
    'LR' / Vector2D,
    'texUL' / Vector2D,
    'texLR' / Vector2D,
)

DetailObjectLump_t = Struct(
    'origin' / Vector,
    'angles' / QAngle,
    'detailModel' / Int16ul,
    'leaf' / Int16ul,
    'lighting' / ColorRGBExp32,
    'lightStyles' / Int32sl,
    'lightStyleCount' / Int8ul,
    'swayAmount' / Int8ul,
    'shapeAngle' / Int8ul,
    'shapeSize' / Int8ul,
    'orientation' / DetailPropOrientation_t,
    'padding2' / Int8ul[3],
    'type' / DetailPropType_t,
    'padding3' / Int8ul[3],
    'scale' / Float32l
)

DetailPropLump_t = Struct(
    'dictLump' / PrefixedArray(Int32sl,
                               PaddedString(DETAIL_NAME_LENGTH, 'ascii')),
    'spritesLump' / PrefixedArray(Int32sl, DetailSpriteLump_t),
    'objectLump' / PrefixedArray(Int32sl, DetailObjectLump_t)
)


@lump_struct
@lump_version(4)
def lump_prpd(header, profile=None):
    return DetailPropLump_t



"""
Lump 28 - Phys Disp
===================

This lump contains a single :any:`physdisps`. 'registry' and 'bytestreams' are of equal length. the length of each bytestream is determined by the corresponding length stated in the registry. The format of the bytestream is not currently known/implemented and will return raw bytes.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA #402

streamLength = Struct('length' / Byte, 'unused' / Byte)

physdisps = Struct(
    'registry' / PrefixedArray(Int16ul, streamLength),
    'bytestreams' / Array(lambda ctx: len(ctx.registry),
                          Bytes(lambda ctx: ctx.registry[ctx._index].length))
)


@lump_version(0)
def lump_28(header, profile=None):
    return physdisps



"""
Lump 45 - Overlays
==================

This lump contains an array of :any:`doverlay_t`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA #402

doverlay_t = Struct(
    'id' / Int32sl,
    'texInfo' / Int16sl * "index into :ref:`lump 6<lump_6>`",
    'faceCountAndRenderOrder' / Int16ul,
    'faces' / Int32sl[OVERLAY_BSP_FACE_COUNT],
    'u' / Float32l[2],
    'v' / Float32l[2],
    'UVPoints' / Vector[4],
    'origin' / Vector,
    'basisNormal' / Vector
)


@lump_array
@lump_version(0)
def lump_45(header, profile=None):
    return doverlay_t



"""
Lump 53 - Lighting HDR
======================

This lump contains an array of :any:`ColorRGBExp32`. They represent individual luxels.
It is the HDR version of :ref:`lump 8<lump_8>`.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from valvebsp.structs.common import *  # NOQA: #402


@lump_array
@lump_version([0, 1])
def lump_53(header, profile=None):
    return ColorRGBExp32



"""
Common Structures
=================

These are reoccuring data structures that you'll see reused throughout Bsp.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: E402
from valvebsp.constants import *  # NOQA: E402
from valvebsp.exceptions import *  # NOQA: #402

import collections
try:
    Iterable = collections.abc.Iterable
except AttributeError:
    Iterable = collections.Iterable


Vector = Struct('x' / Float32l, 'y' / Float32l, 'z' / Float32l)


Vector2D = Struct('x' / Float32l, 'y' / Float32l)
Vector4D = Struct('x' / Float32l, 'y' / Float32l,
                  'z' / Float32l, 'w' / Float32l)
RadianEuler = Struct('x' / Float32l, 'y' / Float32l, 'z' / Float32l)
QAngle = Struct('x' / Float32l, 'y' / Float32l, 'z' / Float32l)
Quaternion = Struct('x' / Float32l, 'y' / Float32l,
                    'z' / Float32l, 'w' / Float32l)
color32 = Struct('r' / Int8ul, 'g' / Int8ul, 'b' / Int8ul, 'a' / Int8ul)
ColorRGBExp32 = Struct('r' / Byte, 'g' / Byte, 'b' /
                       Byte, 'exponent' / Int8sl)
CompressedLightCube = Array(6, ColorRGBExp32)


def lump_version(versions):
    # helper for lump version validation
    def decorator(func):
        def wrapper(*args, **kwargs):
            requested_version = args[0].version
            supported_versions = versions
            if not isinstance(versions, Iterable):
                supported_versions = [versions]
            if requested_version not in supported_versions:
                raise LumpVersionUnsupportedError(requested_version)
            return func(*args, **kwargs)
        return wrapper
    return decorator


def lump_raw(func):
    # helper for raw or unsupported lumps
    def wrapper(*args, **kwargs):
        func(*args, **kwargs)
        return GreedyBytes
    return wrapper


def lump_struct(func):
    # helper for struct based lumps
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs).compile()
    return wrapper


def lump_array(func):
    # helper for array based lumps
    def wrapper(*args, **kwargs):
        return GreedyRange(func(*args, **kwargs).compile())
    return wrapper


def lump_prefixed_array(prefix):
    # helper for prefixed array based lumps
    def decorator(func):
        def wrapper(*args, **kwargs):
            return PrefixedArray(prefix, func(*args, **kwargs))
        return wrapper
    return decorator



"""
Lump 58 - Faces HDR
===================

This lump contains an array of :any:`dface_t`. It is the HDR version of :ref:`lump_8<lump_8>`.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from valvebsp.structs.common import *  # NOQA #402
from valvebsp.structs.lump_7 import dface_t  # NOQA #402


@lump_array
@lump_version([0, 1])
def lump_58(header, profile=None):
    return dface_t



"""
Lump 4 - Visibility
===================

|lump_not_implemented|
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402

'''
dvis_t = Struct(
    'numClusters' / Int32sl,
    'bitofs' / Int32sl[this.numclusters][2]
)
'''


@lump_raw
@lump_version(0)
def lump_4(header, profile=None):
    return



"""
Header
======

The BSP file header is defined as a single :any:`dheader_t` (dheader_t variants are determined by :ref:`profile<profiles>`)
"""


from __future__ import division
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
from builtins import int
from future import standard_library
standard_library.install_aliases()


from construct import *	 # NOQA: #402
from valvebsp.constants import *  # NOQA: #402

lump_t_TF2 = Struct(
    'fileofs' / Int32sl,
    'filelen' / Int32sl,
    'version' / Int32sl,
    'uncompressedSize' / Int32sl
)

lump_t_L4D2 = Struct(
    'version' / Int32sl,
    'fileofs' / Int32sl,
    'filelen' / Int32sl,
    'fourCC' / Byte[4]
)

lump_t = Struct(
    'fileofs' / Int32sl,
    'filelen' / Int32sl,
    'version' / Int32sl,
    'fourCC' / Byte[4]
)

dheader_t_TITAN = Struct(
    'ident' / Const(b'VBSP'),
    'version' / Int32sl,
    'mapRevision' / Default(Int32sl, 0),
    'lump_count' / Const(127, Int32sl),
    'lump_t' / lump_t[this.lump_count],
)

dheader_t_TF2 = Struct(
    'ident' / Const(b'VBSP'),
    'version' / Int32sl,
    'lump_t' / lump_t_TF2[HEADER_LUMPS],
    'mapRevision' / Default(Int32sl, 0)
)

dheader_t_L4D2 = Struct(
    'ident' / Const(b'VBSP'),
    'version' / Int32sl,
    'lump_t' / lump_t_L4D2[HEADER_LUMPS],
    'mapRevision' / Default(Int32sl, 0)
)

dheader_t = Struct(
    'ident' / Const(b'VBSP'),
    'version' / Int32sl,
    'lump_t' / lump_t[HEADER_LUMPS],
    'mapRevision' / Default(Int32sl, 0)
)


def header(profile=None):
    if profile == TEAMFORTRESS2:
        return dheader_t_TF2
    elif profile in [LEFT4DEAD2, CONTAGION]:
        return dheader_t_L4D2
    elif profile == TITANFALL:
        return dheader_t_TITAN
    else:
        return dheader_t



"""
Lump 9 - Occlusion
==================

This lump contains a single :any:`doccluder_t`, regrouping multiple :any:`doccluderdata_t` and :any:`doccluderpolydata_t`
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402
from valvebsp.structs.flags import doccluderdata_flags32  # NOQA: #402


doccluderdata_t = Struct(
    'flags' / doccluderdata_flags32,
    'firstPoly' / Int32sl,
    'polyCount' / Int32sl,
    'mins' / Vector,
    'maxs' / Vector,
    'area' / Int32sl,
)

doccluderpolydata_t = Struct(
    'firstVertexIndex' / Int32sl,
    'vertexCount' / Int32sl,
    'planeNum' / Int32sl,
)

doccluder_t = Struct(
    'count' / Int32sl,
    'data' / doccluderdata_t[this.count],
    'polyDataCount' / Int32sl,
    'polyData' / doccluderpolydata_t[this.polyDataCount],
    'vertexIndexCount' / Int32sl,
    'vertexIndices' / Int32sl[this.vertexIndexCount],
)


@lump_struct
@lump_version(2)
def lump_9(header, profile=None):
    return doccluder_t



"""
Lump 14 - Models
================

This lump contains an array of :any:`dmodel_t`.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402


dmodel_t = Struct(
    'mins' / Vector * 'lower point of the bounding box',
    'maxs' / Vector * 'higher point the bounding box',
    'origin' / Vector,
    'headnode' / Int32sl * 'index into :ref:`lump 5<lump_5>`',
    'firstface' / Int32sl * 'index into :ref:`lump 7<lump_7>`',
    'numfaces' / Int32sl
)


@lump_array
@lump_version(0)
def lump_14(header, profile=None):
    return dmodel_t



"""
Lump 42 - Cubemaps
==================

This lump contains an array of :any:`dcubemapsample_t`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA #402

dcubemapsample_t = Struct(
    'origin' / Int32sl[3],
    'size' / Int32sl
)


@lump_array
@lump_version(0)
def lump_42(header, profile=None):
    return dcubemapsample_t



"""
Lump 33 - Disp Verts
====================

This lump contains an array of :any:`CDispVert`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA #402

CDispVert = Struct(
    'vector' / Vector,
    'dist' / Float32l,
    'alpha' / Float32l,
)


@lump_array
@lump_version(0)
def lump_33(header, profile=None):
    return CDispVert



"""
Lump 51 - Lightmap Pages ...or Leaf Ambient Index HDR
=====================================================

This lump contains an array of :any:`dleafambientindex_t`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402

dleafambientindex_t = Struct(  # matches dleaf_t
    'ambientSampleCount' / Int16ul,
    'firstAmbientSample' / Int16ul * \
    "index into :ref:`lump 55<lump_55>` or :ref:`lump 56<lump_56>`",
)


@lump_array
@lump_version([0, 1])
def lump_51(header, profile=None):
    return dleafambientindex_t



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from valvebsp.structs.bsp_header import header  # NOQA: #402
from valvebsp.structs.lump_0 import lump_0  # NOQA: #402
from valvebsp.structs.lump_1 import lump_1  # NOQA: #402
from valvebsp.structs.lump_2 import lump_2  # NOQA: #402
from valvebsp.structs.lump_3 import lump_3  # NOQA: #402
from valvebsp.structs.lump_4 import lump_4  # NOQA: #402
from valvebsp.structs.lump_5 import lump_5  # NOQA: #402
from valvebsp.structs.lump_6 import lump_6  # NOQA: #402
from valvebsp.structs.lump_7 import lump_7  # NOQA: #402
from valvebsp.structs.lump_8 import lump_8  # NOQA: #402
from valvebsp.structs.lump_9 import lump_9  # NOQA: #402
from valvebsp.structs.lump_10 import lump_10  # NOQA: #402
from valvebsp.structs.lump_11 import lump_11  # NOQA: #402
from valvebsp.structs.lump_12 import lump_12  # NOQA: #402
from valvebsp.structs.lump_13 import lump_13  # NOQA: #402
from valvebsp.structs.lump_14 import lump_14  # NOQA: #402
from valvebsp.structs.lump_15 import lump_15  # NOQA: #402
from valvebsp.structs.lump_16 import lump_16  # NOQA: #402
from valvebsp.structs.lump_17 import lump_17  # NOQA: #402
from valvebsp.structs.lump_18 import lump_18  # NOQA: #402
from valvebsp.structs.lump_19 import lump_19  # NOQA: #402
from valvebsp.structs.lump_20 import lump_20  # NOQA: #402
from valvebsp.structs.lump_21 import lump_21  # NOQA: #402
from valvebsp.structs.lump_22 import lump_22  # NOQA: #402
from valvebsp.structs.lump_23 import lump_23  # NOQA: #402
from valvebsp.structs.lump_24 import lump_24  # NOQA: #402
from valvebsp.structs.lump_25 import lump_25  # NOQA: #402
from valvebsp.structs.lump_26 import lump_26  # NOQA: #402
from valvebsp.structs.lump_27 import lump_27  # NOQA: #402
from valvebsp.structs.lump_28 import lump_28  # NOQA: #402
from valvebsp.structs.lump_29 import lump_29  # NOQA: #402
from valvebsp.structs.lump_30 import lump_30  # NOQA: #402
from valvebsp.structs.lump_31 import lump_31  # NOQA: #402
from valvebsp.structs.lump_32 import lump_32  # NOQA: #402
from valvebsp.structs.lump_33 import lump_33  # NOQA: #402
from valvebsp.structs.lump_34 import lump_34  # NOQA: #402
from valvebsp.structs.lump_35 import lump_35  # NOQA: #402
from valvebsp.structs.lump_36 import lump_36  # NOQA: #402
from valvebsp.structs.lump_37 import lump_37  # NOQA: #402
from valvebsp.structs.lump_38 import lump_38  # NOQA: #402
from valvebsp.structs.lump_39 import lump_39  # NOQA: #402
from valvebsp.structs.lump_40 import lump_40  # NOQA: #402
from valvebsp.structs.lump_41 import lump_41  # NOQA: #402
from valvebsp.structs.lump_42 import lump_42  # NOQA: #402
from valvebsp.structs.lump_43 import lump_43  # NOQA: #402
from valvebsp.structs.lump_44 import lump_44  # NOQA: #402
from valvebsp.structs.lump_45 import lump_45  # NOQA: #402
from valvebsp.structs.lump_46 import lump_46  # NOQA: #402
from valvebsp.structs.lump_47 import lump_47  # NOQA: #402
from valvebsp.structs.lump_48 import lump_48  # NOQA: #402
from valvebsp.structs.lump_49 import lump_49  # NOQA: #402
from valvebsp.structs.lump_50 import lump_50  # NOQA: #402
from valvebsp.structs.lump_51 import lump_51  # NOQA: #402
from valvebsp.structs.lump_52 import lump_52  # NOQA: #402
from valvebsp.structs.lump_53 import lump_53  # NOQA: #402
from valvebsp.structs.lump_54 import lump_54  # NOQA: #402
from valvebsp.structs.lump_55 import lump_55  # NOQA: #402
from valvebsp.structs.lump_56 import lump_56  # NOQA: #402
from valvebsp.structs.lump_57 import lump_57  # NOQA: #402
from valvebsp.structs.lump_58 import lump_58  # NOQA: #402
from valvebsp.structs.lump_59 import lump_59  # NOQA: #402
from valvebsp.structs.lump_60 import lump_60  # NOQA: #402
from valvebsp.structs.lump_61 import lump_61  # NOQA: #402
from valvebsp.structs.lump_62 import lump_62  # NOQA: #402
from valvebsp.structs.lump_63 import lump_63  # NOQA: #402
from valvebsp.structs.lump_60 import lump_60  # NOQA: #402
from valvebsp.structs.lump_61 import lump_61  # NOQA: #402
from valvebsp.structs.lump_62 import lump_62  # NOQA: #402
from valvebsp.structs.lump_63 import lump_63  # NOQA: #402
from valvebsp.structs.lump_prps import lump_prps  # NOQA: #402
from valvebsp.structs.lump_prpd import lump_prpd  # NOQA: #402
from valvebsp.structs.lump_tlpd import lump_tlpd  # NOQA: #402
from valvebsp.structs.lump_hlpd import lump_hlpd  # NOQA: #402



"""
Lump 20 - Areas
===============

This lump contains an array of :any:`darea_t`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402

darea_t = Struct(
    'numAreaPortals' / Int32sl,
    'firstAreaPortal' / Int32sl * "index into :ref:`lump 21<lump_21>`"
)


@lump_array
@lump_version(0)
def lump_20(header, profile=None):
    return darea_t



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

LUMP_ENTITIES = 0  #:
LUMP_PLANES = 1  #:
LUMP_TEXDATA = 2  #:
LUMP_VERTEXES = 3  #:
LUMP_VISIBILITY = 4  #:
LUMP_NODES = 5  #:
LUMP_TEXINFO = 6  #:
LUMP_FACES = 7  #:
LUMP_LIGHTING = 8  #:
LUMP_OCCLUSION = 9  #:
LUMP_LEAFS = 10  #:
LUMP_FACEIDS = 11  #:
LUMP_EDGES = 12  #:
LUMP_SURFEDGES = 13  #:
LUMP_MODELS = 14  #:
LUMP_WORLDLIGHTS = 15  #:
LUMP_LEAFFACES = 16  #:
LUMP_LEAFBRUSHES = 17  #:
LUMP_BRUSHES = 18  #:
LUMP_BRUSHSIDES = 19  #:
LUMP_AREAS = 20  #:
LUMP_AREAPORTALS = 21  #:
LUMP_UNUSED0 = 22  #:
LUMP_UNUSED1 = 23  #:
LUMP_UNUSED2 = 24  #:
LUMP_UNUSED3 = 25  #:
LUMP_DISPINFO = 26  #:
LUMP_ORIGINALFACES = 27  #:
LUMP_PHYSDISP = 28  #:
LUMP_PHYSCOLLIDE = 29  #:
LUMP_VERTNORMALS = 30  #:
LUMP_VERTNORMALINDICES = 31  #:
LUMP_DISP_LIGHTMAP_ALPHAS = 32  #:
LUMP_DISP_VERTS = 33  #:
LUMP_DISP_LIGHTMAP_SAMPLE_POSITIONS = 34  #:
LUMP_GAME_LUMP = 35  #:
LUMP_LEAFWATERDATA = 36  #:
LUMP_PRIMITIVES = 37  #:
LUMP_PRIMVERTS = 38  #:
LUMP_PRIMINDICES = 39  #:
LUMP_PAKFILE = 40  #:
LUMP_CLIPPORTALVERTS = 41  #:
LUMP_CUBEMAPS = 42  #:
LUMP_TEXDATA_STRING_DATA = 43  #:
LUMP_TEXDATA_STRING_TABLE = 44  #:
LUMP_OVERLAYS = 45  #:
LUMP_LEAFMINDISTTOWATER = 46  #:
LUMP_FACE_MACRO_TEXTURE_INFO = 47  #:
LUMP_DISP_TRIS = 48  #:
LUMP_PHYSCOLLIDESURFACE = 49  #:
LUMP_WATEROVERLAYS = 50  #:
LUMP_LEAF_AMBIENT_INDEX_HDR = 51  #:
LUMP_LEAF_AMBIENT_INDEX = 52  #:
LUMP_LIGHTING_HDR = 53  #:
LUMP_WORLDLIGHTS_HDR = 54  #:
LUMP_LEAF_AMBIENT_LIGHTING_HDR = 55  #:
LUMP_LEAF_AMBIENT_LIGHTING = 56  #:
LUMP_XZIPPAKFILE = 57  #:
LUMP_FACES_HDR = 58  #:
LUMP_MAP_FLAGS = 59  #:
LUMP_OVERLAY_FADES = 60  #:
LUMP_OVERLAY_SYSTEM_LEVELS = 61  #:
LUMP_PHYSLEVEL = 62  #:
LUMP_DISP_MULTIBLEND = 63  #:



"""
Lump 52 - Lightmap Page Infos ...or Leaf Ambient Index
======================================================

This lump contains an array of :any:`dleafambientindex_t` (:ref:`as seen in lump_51<lump_51>`).
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from valvebsp.structs.common import *  # NOQA #402
from valvebsp.structs.lump_51 import dleafambientindex_t  # NOQA #402


@lump_array
@lump_version([0, 1])
def lump_52(header, profile=None):
    return dleafambientindex_t



"""
Lump prps - Prop Static
=======================

This lump is defined as a single StaticPropLump_t that may come in different versions, in accordance with the gamelump.`
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402
from valvebsp.structs.flags import prps_flags8  # NOQA: #402


StaticPropV4_t = Struct(
    'origin' / Vector,
    'angles' / QAngle,

    'propType' / Int16ul * 'refers to prps.dict_lump',
    'firstLeaf' / Int16ul * 'refers to prps.leaf_lump',
    'leafCount' / Int16ul,
    'solid' / Int8ul,
    'flags' / prps_flags8,
    'skin' / Int32sl,

    'fadeMinDist' / Float32l,
    'fadeMaxDist' / Float32l,

    'lightingOrigin' / Vector,
)

StaticPropV5_t = Struct(
    # hl, hl2, p1, css
    'origin' / Vector,
    'angles' / QAngle,

    'propType' / Int16ul * 'refers to prps.dict_lump',
    'firstLeaf' / Int16ul * 'refers to prps.leaf_lump',
    'leafCount' / Int16ul,
    'solid' / Int8ul,
    'flags' / prps_flags8,
    'skin' / Int32sl,

    'fadeMinDist' / Float32l,
    'fadeMaxDist' / Float32l,

    'lightingOrigin' / Vector,
    'forcedFadeScale' / Float32l,
)

StaticPropV6_t = Struct(
    # p1, css, dod
    'origin' / Vector,
    'angles' / QAngle,

    'propType' / Int16ul * 'refers to prps.dict_lump',
    'firstLeaf' / Int16ul * 'refers to prps.leaf_lump',
    'leafCount' / Int16ul,
    'solid' / Int8ul,
    'flags' / prps_flags8,
    'skin' / Int32sl,

    'FadeMinDist' / Float32l,
    'FadeMaxDist' / Float32l,

    'lightingOrigin' / Vector,
    'forcedFadeScale' / Float32l,

    'minDXLevel' / Int16ul,
    'maxDXLevel' / Int16ul,
)

StaticPropV7_t = Struct(
    # zeno
    'origin' / Vector,
    'angles' / QAngle,

    'propType' / Int16ul * 'refers to prps.dict_lump',
    'firstLeaf' / Int16ul * 'refers to prps.leaf_lump',
    'leafCount' / Int16ul,
    'solid' / Int8ul,
    'flags' / prps_flags8,
    'skin' / Int32sl,

    'FadeMinDist' / Float32l,
    'FadeMaxDist' / Float32l,

    'lightingOrigin' / Vector,
    'forcedFadeScale' / Float32l,

    'minDXLevel' / Int16ul,
    'maxDXLevel' / Int16ul,

    'diffuseModulation' / color32,
)

StaticPropV8_t = Struct(
    'origin' / Vector,
    'angles' / QAngle,

    'propType' / Int16ul * 'refers to prps.dict_lump',
    'firstLeaf' / Int16ul * 'refers to prps.leaf_lump',
    'leafCount' / Int16ul,
    'solid' / Int8ul,
    'flags' / prps_flags8,
    'skin' / Int32sl,

    'FadeMinDist' / Float32l,
    'FadeMaxDist' / Float32l,

    'lightingOrigin' / Vector,
    'forcedFadeScale' / Float32l,

    'minCPULevel' / Int8ul,
    'maxCPULevel' / Int8ul,
    'minGPULevel' / Int8ul,
    'maxGPULevel' / Int8ul,

    'diffuseModulation' / color32,
)


StaticPropV9_t = Struct(
    # portal2
    'origin' / Vector,
    'angles' / QAngle,

    'propType' / Int16ul * 'refers to prps.dict_lump',
    'firstLeaf' / Int16ul * 'refers to prps.leaf_lump',
    'leafCount' / Int16ul,
    'solid' / Int8ul,
    'flags' / prps_flags8,
    'skin' / Int32sl,

    'FadeMinDist' / Float32l,
    'FadeMaxDist' / Float32l,

    'lightingOrigin' / Vector,
    'forcedFadeScale' / Float32l,

    'minCPULevel' / Int8ul,
    'maxCPULevel' / Int8ul,
    'minGPULevel' / Int8ul,
    'maxGPULevel' / Int8ul,

    'diffuseModulation' / color32,

    'disableX360' / Aligned(4, Flag)
)


StaticPropV10_t = Aligned(4, Struct(
    # tf2
    'origin' / Vector,
    'angles' / QAngle,

    'propType' / Int16ul * 'refers to prps.dict_lump',
    'firstLeaf' / Int16ul * 'refers to prps.leaf_lump',
    'leafCount' / Int16ul,

    'solid' / Int8ul,
    'flags' / prps_flags8,
    'skin' / Int32sl,

    'FadeMinDist' / Float32l,
    'FadeMaxDist' / Float32l,

    'lightingOrigin' / Vector,
    'forcedFadeScale' / Float32l,

    'minCPULevel' / Int8ul,
    'maxCPULevel' / Int8ul,
    'minGPULevel' / Int8ul,
    'maxGPULevel' / Int8ul,

    'diffuseModulation' / color32,

    'flagsEx' / Int32ul,
))

StaticPropV11_t = Aligned(4, Struct(
    # csgo
    'origin' / Vector,
    'angles' / QAngle,

    'propType' / Int16ul * 'refers to prps.dict_lump',
    'firstLeaf' / Int16ul * 'refers to prps.leaf_lump',
    'leafCount' / Int16ul,

    'solid' / Int8ul,
    'flags' / prps_flags8,
    'skin' / Int32sl,

    'FadeMinDist' / Float32l,
    'FadeMaxDist' / Float32l,

    'lightingOrigin' / Vector,
    'forcedFadeScale' / Float32l,

    'minCPULevel' / Int8ul,
    'maxCPULevel' / Int8ul,
    'minGPULevel' / Int8ul,
    'maxGPULevel' / Int8ul,

    'diffuseModulation' / color32,

    'disableX360' / Aligned(4, Flag),

    'flagsEx' / Int32ul,

    'uniformScale' / Float32l,
))


@lump_version([4, 5, 6, 7, 8, 9, 10, 11])
def lump_prps(header, profile=None):
    if header.version == 11:
        StaticProp_t = StaticPropV11_t
    elif header.version == 10:
        StaticProp_t = StaticPropV10_t
    elif header.version == 9:
        StaticProp_t = StaticPropV9_t
    elif header.version == 8:
        StaticProp_t = StaticPropV8_t
    elif header.version == 7:
        StaticProp_t = StaticPropV7_t
    elif header.version == 6:
        StaticProp_t = StaticPropV6_t
    elif header.version == 5:
        StaticProp_t = StaticPropV5_t
    elif header.version == 4:
        StaticProp_t = StaticPropV4_t

    if profile == ZENOCLASH:
        StaticPropLump_t = Struct(
            'dictLump' / PrefixedArray(Int32sl, PaddedString(
                STATIC_PROP_NAME_LENGTH, 'ascii')),
            'dictLump2' / PrefixedArray(Int32sl, PaddedString(
                STATIC_PROP_NAME_LENGTH, 'ascii')),
            'leafLump' / PrefixedArray(Int32sl, Int16ul),
            'objectLump' / PrefixedArray(Int32sl, StaticProp_t))
    else:
        StaticPropLump_t = Struct(
            'dictLump' / PrefixedArray(Int32sl, PaddedString(
                STATIC_PROP_NAME_LENGTH, 'ascii')),
            'leafLump' / PrefixedArray(Int32sl, Int16ul),
            'objectLump' / PrefixedArray(Int32sl, StaticProp_t))

    return StaticPropLump_t



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import open
from builtins import range
from builtins import str
from future import standard_library
standard_library.install_aliases()

import collections
from shutil import copyfile  # NOQA: #402

from construct import *  # NOQA: #402
import valvebsp.structs.bsp as BSP  # NOQA: #402

try:
    collectionsAbc = collections.abc
except AttributeError:
    collectionsAbc = collections


class LumpUnsupportedError(IndexError):
    pass


def byte_align(filelen, align=4):
    if (filelen % align):
        filelen += (align - filelen % align)
    return filelen


class Bsp(collectionsAbc.MutableMapping):
    """Contains all the data from a Bsp file"""

    def __init__(self, path=None, profile=None):
        """Creates an empty instance of Bsp.

        :param path: A path to an existing bsp file.
        :type path: str
        :param profile: A profile name corresponding to a specific game.
            :ref:`See profiles page for appropriate values<profiles>`
        :type profile: str, optional"""

        self.source_path = path
        self.profile = profile

        self.header = None
        self.lumps = {}

        # note: file might not be found,
        # profile might not exist
        if self.source_path:
            with open(path, 'rb') as f:
                header_struct = BSP.header(self.profile)
                self.header = header_struct.parse_stream(f)

    def save(self, destination=None):
        """Saves the current instance of the Bsp. Overwrites original bsp
        file if no destination is provided.

        Note that the bsp is lazy loaded. When overwritting, only loaded
        lumps will be written. When saving to a new destination,
        The original bsp will be copied and loaded lumps will be
        written on top.

        :param destination: A path (directory + filename) to determine
            where to save the bsp file.
        :type destination: str, optional
        """

        dest = destination or self.source_path
        header_struct = BSP.header(self.profile)

        if not dest:
            # save location is invalid
            raise FileNotFoundError

        elif not self.source_path:
            # creating a new file from nothing
            try:
                d = open(dest, 'wb')
                self._build_stream(header_struct, self.header, d)
            except:
                raise FileNotFoundError

        elif dest == self.source_path:
            # overwritting original file
            try:
                d = open(dest, 'rb+')
            except:
                raise FileNotFoundError
        else:
            # source and dest are different
            s = open(self.source_path, 'rb')
            d = open(dest, 'wb+')
            d.write(s.read())
            s.close()

        self[35]  # loading gamelump for potential updates

        for key in self.lumps.keys():
            val = self.lumps[key]

            lump_header = self._get_lump_header(key)
            lump_struct = self._get_lump_struct(key)

            data = lump_struct.build(val)

            if key == 35:
                continue

            if len(data) == lump_header.filelen:
                d.seek(lump_header.fileofs)
                d.write(data)
            else:
                tail_from = lump_header.fileofs + \
                    byte_align(lump_header.filelen)
                tail_to = lump_header.fileofs + \
                    byte_align(len(data))

                # read data to be offset
                d.seek(tail_from)
                tail = d.read()

                # write lump_data
                d.seek(lump_header.fileofs)
                d.write(data)

                # write data to be offset
                d.seek(tail_to)

                d.write(tail)
                d.truncate()

                # update headers
                difference = tail_to - tail_from

                lump_header.filelen = len(data)

                for lump_t in self.header.lump_t:
                    if lump_header.fileofs < lump_t.fileofs:
                        lump_t.fileofs = lump_t.fileofs + difference

                for g_lump in self[35].gamelump:
                    if lump_header.fileofs < g_lump.fileofs:
                        g_lump.fileofs = g_lump.fileofs + difference

        # write lump headers
        header_data = header_struct.build(self.header)
        d.seek(0)
        d.write(header_data)

        # write game headers
        glump_header = self._get_lump_header(35)
        glump_struct = self._get_lump_struct(35)
        glump_data = glump_struct.build(self[35])
        d.seek(glump_header.fileofs)
        d.write(glump_data)

        d.close()

    def _parse_stream(self, struct, f, **kwargs):
        kwargs['bspHeader'] = self.header
        return struct.parse_stream(f, **kwargs)

    def _parse_file(self, struct, **kwargs):
        kwargs['bspHeader'] = self.header
        return struct.parse_file(self.source_path, **kwargs)

    def _build_stream(self, struct, data, f, **kwargs):
        kwargs['bspHeader'] = self.header
        if not data:
            return
        return struct.build_stream(data, f, **kwargs)

    def _build_file(self, struct, data, **kwargs):
        if not struct:
            return
        kwargs['bspHeader'] = self.header
        return struct.build_file(data, self.source_path, **kwargs)

    def _get_lump_header(self, index):
        if isinstance(index, str):  # It's a gamelump id
            for h in self[35].gamelump:
                if h.id == index:
                    return h
        elif isinstance(index, int):  # It's a lump number
            if index in range(64) and self.header:
                return self.header.lump_t[index]

        raise LumpUnsupportedError('Invalid Lump ID (' + str(index) + ')')

    def _get_lump_struct(self, index):
        lump_header = self._get_lump_header(index)
        lump_fn = getattr(BSP, 'lump_' + str(index))
        lump_struct = lump_fn(lump_header, self.profile)
        return lump_struct

    def __getitem__(self, index):
        """Provides data at the specified lump index. It is used for both
        bsp lumps (0, 1, 2, ...63) and game lumps ('prps', 'prpd', 'tlpd'...)

        :param index: The index of the lump.
        :type index: str, int
        """

        if index in self.lumps and self.lumps[index]:
            return self.lumps[index]

        lump_header = self._get_lump_header(index)
        lump_struct = self._get_lump_struct(index)

        with open(self.source_path, 'rb') as f:
            f.seek(lump_header.fileofs)
            lump_raw = f.read(lump_header.filelen)

        lump_data = lump_struct.parse(lump_raw)
        self.lumps[index] = lump_data

        return self.lumps[index]

    def __setitem__(self, index, value):
        if index not in range(0, 64) and \
           index not in ['prps', 'prpd', 'tlpd', 'hlpd']:
            raise LumpUnsupportedError(index)

        self.lumps[index] = value

    def __delitem__(self, key):
        if key in self.lumps:
            del self.lumps[self.__keytransform__(key)]

    def __iter__(self):
        return self.item(self.lumps)

    def __len__(self, index):
        return 64

    def __keytransform__(self, key):
        return key



"""
Lump 59 - Map Flags
===================

This lump contains a single :any:`flags_t`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA #402
from valvebsp.structs.flags import levelflags_flags32  # NOQA: #402

flags_t = Struct('levelFlags' / levelflags_flags32)


@lump_struct
@lump_version(0)
def lump_59(header, profile=None):
    return flags_t



"""
Lump 56 - Leaf Ambient Lighting
===============================

This lump contains an array of :any:`dleafambientlighting_t`.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA #402

dleafambientlighting_t = Aligned(4, Struct(
    'cube' / CompressedLightCube,
    'x' / Byte,
    'y' / Byte,
    'z' / Byte
))


@lump_array
def lump_56(header, profile=None):
    if header.version == 1:
        return dleafambientlighting_t

    elif header.version == header.filelen:
        # This is obviously a mistake in the bsp header, let's assume v1
        # (seen in css map once)
        return dleafambientlighting_t

    else:
        raise LumpVersionUnsupportedError(header.version)



"""
Lump 5 - Nodes
==============

This lump contains an array of :any:`dnode_t`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA #402

dnode_t = Aligned(4, Struct(
    'planeNum' / Int32sl,
    'children' / Int32sl[2],
    'mins' / Int16sl[3],
    'maxs' / Int16sl[3],
    'firstFace' / Int16ul * 'index into :ref:`lump 7<lump_7>`',
    'numFaces' / Int16ul,
    'area' / Int16sl,
))


@lump_array
@lump_version(0)
def lump_5(header, profile=None):
    return dnode_t



"""
Lump 27 - Original Faces
========================

This lump contains an array of :any:`dface_t` (:ref:`as seen in lump_7<lump_7>`).
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA #402
from valvebsp.structs.lump_7 import dface_t  # NOQA #402


@lump_array
@lump_version(0)
def lump_27(header, profile=None):
    return dface_t



"""
Lump 40 - Pakfile
=================

This lump contains raw data. The data can be saved directly to the disc as a zip file.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from valvebsp.structs.common import *  # NOQA #402


@lump_raw
@lump_version(0)
def lump_40(header, profile=None):
    return



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402

contents_flags32 = FlagsEnum(Int32ul,
                             CONTENTS_SOLID=1,
                             CONTENTS_WINDOW=2,
                             CONTENTS_AUX=4,
                             CONTENTS_GRATE=8,
                             CONTENTS_SLIME=16,
                             CONTENTS_WATER=32,
                             CONTENTS_BLOCKLOS=64,
                             CONTENTS_OPAQUE=128,
                             CONTENTS_TESTFOGVOLUME=256,
                             CONTENTS_UNUSED=512,
                             CONTENTS_UNUSED6=1024,
                             CONTENTS_TEAM1=2048,
                             CONTENTS_TEAM2=4096,
                             CONTENTS_IGNORE_NODRAW_OPAQUE=8192,
                             CONTENTS_MOVEABLE=16384,
                             CONTENTS_AREAPORTAL=32768,
                             CONTENTS_PLAYERCLIP=65536,
                             CONTENTS_MONSTERCLIP=131072,
                             CONTENTS_CURRENT_0=262144,
                             CONTENTS_CURRENT_90=524288,
                             CONTENTS_CURRENT_180=1048576,
                             CONTENTS_CURRENT_270=2097152,
                             CONTENTS_CURRENT_UP=4194304,
                             CONTENTS_CURRENT_DOWN=8388608,
                             CONTENTS_ORIGIN=16777216,
                             CONTENTS_MONSTER=33554432,
                             CONTENTS_DEBRIS=67108864,
                             CONTENTS_DETAIL=134217728,
                             CONTENTS_TRANSLUCENT=268435456,
                             CONTENTS_LADDER=536870912,
                             CONTENTS_HITBOX=1073741824)
dispinfo_flags8 = FlagsEnum(Int8ul,
                            COLL_UNUSED=1,
                            COLL_NOPHYS=2,
                            COLL_NOHULL=4,
                            COLL_NORAY=8)
texinfo_flags32 = FlagsEnum(Int32ul,
                            SURF_LIGHT=1,
                            SURF_SKY2D=2,
                            SURF_SKY=4,
                            SURF_WARP=8,
                            SURF_TRANS=16,
                            SURF_NOPORTAL=32,
                            SURF_TRIGGER=64,
                            SURF_NODRAW=128,
                            SURF_HINT=256,
                            SURF_SKIP=512,
                            SURF_NOLIGHT=1024,
                            SURF_BUMPLIGHT=2048,
                            SURF_NOSHADOWS=4096,
                            SURF_NODECALS=8192,
                            SURF_NOCHOP=16384,
                            SURF_HITBOX=32768)

dworldlight_flags32 = FlagsEnum(Int32sl,
                                DWL_FLAGS_INAMBIENTCUBE=0,
                                DWL_FLAGS_CASTENTITYSHADOWS=1)

dgamelump_flags16 = FlagsEnum(Int16ul,
                              GAMELUMPFLAG_COMPRESSED=1)

doccluderdata_flags32 = FlagsEnum(Int32sl,
                                  OCCLUDER_FLAGS_INACTIVE=1)

levelflags_flags32 = FlagsEnum(Int32ul,
                               BAKED_STATIC_PROP_LIGHTING_NONHDR=1,
                               BAKED_STATIC_PROP_LIGHTING_HDR=2)

dleaf_flags7 = FlagsEnum(BitsInteger(7),
                         LEAF_FLAGS_SKY=1,
                         LEAF_FLAGS_RADIAL=2,
                         LEAF_FLAGS_SKY2D=4,
                         LEAF_FLAGS_CONTAINS_DETAILOBJECTS=8)

prps_flags8 = FlagsEnum(Int8ul,
                        STATIC_PROP_FLAG_FADES=1,
                        STATIC_PROP_USE_LIGHTING_ORIGIN=2,
                        STATIC_PROP_NO_DRAW=4,
                        STATIC_PROP_IGNORE_NORMALS=8,
                        STATIC_PROP_NO_SHADOW=16,
                        STATIC_PROP_SCREEN_SPACE_FADE_OBSOLETE=32,
                        STATIC_PROP_NO_PER_VERTEX_LIGHTING=64,
                        STATIC_PROP_NO_SELF_SHADOWING=128)



"""
Lump 13 - Surfedges
===================

This lump contains an array of :any:`Int<int>`. The absolute value references the edge array (:ref:`lump_12<lump_12>`). Positive values indicate the edge is defined from the first to second vertex. Negative values indicate the edge is defined from second to first vertex.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402


@lump_array
@lump_version(0)
def lump_13(header, profile=None):
    return Int32sl



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from future import standard_library
standard_library.install_aliases()

TEAMFORTRESS2 = 'TF2'  #:
PORTAL2 = 'P2'  #:
LEFT4DEAD2 = 'L4D2'  #:
ALIENSWARM = 'AS'  #:
ZENOCLASH = 'ZC'  #:
VINDICTUS = 'VIN'  #:
CONTAGION = 'CON'  #:
TITANFALL = 'TIT'  #:
THESHIP = 'SHIP'  #:

#: :obj:`dict` :
#: Utility dict listing available profiles.
PROFILES = {
    'TEAMFORTRESS2': TEAMFORTRESS2,
    'PORTAL2': PORTAL2,
    'LEFT4DEAD2': LEFT4DEAD2,
    'ALIENSWARM': ALIENSWARM,
    'ZENOCLASH': ZENOCLASH,
    'VINDICTUS': VINDICTUS,
    'CONTAGION': CONTAGION,
    'TITANFALL': TITANFALL,
    'THESHIP': THESHIP,
}



"""
Lump 26 - Disp Info
===================

This lump contains an array of :any:`ddispinfo_t`
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA #402
from valvebsp.structs.flags import dispinfo_flags8, contents_flags32  # NOQA #402

NeighborSpan = Enum(
    Int8sl,
    CORNER_TO_CORNER=0,
    CORNER_TO_MIDPOINT=1,
    MIDPOINT_TO_CORNER=2
)

NeighborOrientation = Enum(
    Int8sl,
    ORIENTATION_CCW_0=0,
    ORIENTATION_CCW_90=1,
    ORIENTATION_CCW_180=2,
    ORIENTATION_CCW_270=3
)

CDispSubNeighbor = Struct(
    # 'm_iNeighbor' / Int16ul,
    # 'm_NeighborOrientation' / NeighborOrientation,
    # 'm_Span' / NeighborSpan,
    # 'm_NeighborSpan' / NeighborSpan
    # The fields seem to be misordered, this is my best guess
    'span' / NeighborSpan,
    'neighborSpan' / NeighborSpan,
    'neighbor' / Int16ul,
    'unknown' / Byte,

    'neighborOrientation' / NeighborOrientation,
)

CDispNeighbor = Struct(
    'subNeighbors' / CDispSubNeighbor[2],
)

CDispCornerNeighbors = Struct(
    'neighbors' / Int16ul[MAX_DISP_CORNER_NEIGHBORS],
    'nNeighbors' / Int8ul
)

ddispinfo_t = Struct(
    'startPosition' / Vector,
    'dispVertStart' / Int32sl,
    'dispTriStart' / Int32sl,

    'power' / Int32sl,
    'flags' / dispinfo_flags8,
    'minTess' / Int24sl,
    'smoothingAngle' / Float32l,
    'contents' / contents_flags32,

    'mapFace' / Int16ul * "index into :ref:`lump 7<lump_7>` or :ref:`lump 58<lump_58>`",

    'lightmapAlphaStart' / Int32sl * "index into :ref:`lump 32<lump_32>`",
    'lightmapSamplePositionStart' / Int32sl,

    'edgeNeighbors' / CDispNeighbor[4],
    'cornerNeighbors' / CDispCornerNeighbors[4],

    'unknown' / Bytes(6),

    'allowedVerts' / Int32ul[ALLOWEDVERTS_SIZE],
)


@lump_array
@lump_version(0)
def lump_26(header, profile=None):
    return ddispinfo_t



"""
Lump 36 - Leaf Water Data
=========================

This lump contains an array of :any:`dleafwaterdata_t`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA #402

dleafwaterdata_t = Aligned(4, Struct(
    'surfaceZ' / Float32l,
    'minZ' / Float32l,
    'surfaceTexInfoID' / Int16sl,
))


@lump_array
@lump_version(0)
def lump_36(header, profile=None):
    return dleafwaterdata_t



"""
Lump hlpd - Prop Detail Lighting HDR
====================================

This lump is the HDR version of :ref:`lump tlpd<lump_tlpd>`, its structure is identical.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from valvebsp.structs.common import *  # NOQA #402
from valvebsp.structs.lump_tlpd import DetailPropLightStyle_t  # NOQA #402


@lump_prefixed_array(Int32sl)
@lump_version(0)
def lump_hlpd(header, profile=None):
    return DetailPropLightStyle_t



"""
Lump 11 - Face Ids
==================

This lump contains an array of :any:`dfaceid_t`. This array matches the array of dface_t (:ref:`see lump_7<lump_7>`).
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402

dfaceid_t = Struct(
    'hammerFaceID' / Int16ul * "index into :ref:`lump 27<lump_27>`",
)


@lump_array
@lump_version(0)
def lump_11(header, profile=None):
    return dfaceid_t



"""
Lump 44 - Texture Data String Table
===================================

This lump contains an array of :any:`Int<int>`. They represents offsets in :ref:`lump_43<lump_43>`.

Note: Ids pointing to lump 44 should map perfectly to lump 43. It's considered safe to use lump 43 directly instead of 44.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA #402


@lump_array
@lump_version(0)
def lump_44(header, profile=None):
    return Int32sl



"""
Lump 7 - Faces
==============

This lump contains an array of :any:`dface_t`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402

dface_t = Struct(
    'planenum' / Int16ul,
    'side' / Byte,
    'onNode' / Byte,
    'firstedge' / Int32sl * "index into :ref:`lump 13<lump_13>`",
    'numedges' / Int16sl,
    'texinfo' / Int16sl * "index into :ref:`lump 6<lump_6>`",
    'dispinfo' / Int16sl * "index into :ref:`lump 26<lump_26>`",
    'surfaceFogVolumeID' / Int16sl,
    'styles' / Byte[4],
    'lightofs' / Int32sl * "index into :ref:`lump 8<lump_8>`",
    'area' / Float32l,
    'lightmapTextureMinsInLuxels' / Int32sl[2],
    'lightmapTextureSizeInLuxels' / Int32sl[2],
    'origFace' / Int32sl * "index into :ref:`lump 11<lump_11>`",
    'numPrims' / Int16ul,
    'firstPrimID' / Int16ul * "index into :ref:`lump 37<lump_37>`",
    'smoothingGroups' / Int32ul,
)


@lump_array
@lump_version([0, 1])
def lump_7(header, profile=None):
    return dface_t



"""
Lump 55 - Leaf Ambient Lighting HDR
===================================

This lump contains an array of :any:`dleafambientlighting_t`.
It is the HDR version of :ref:`lump 56<lump_56>`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from valvebsp.structs.common import *  # NOQA: #402
from valvebsp.structs.lump_56 import dleafambientlighting_t  # NOQA: #402


@lump_array
def lump_55(header, profile=None):
    if header.version == 1:
        return dleafambientlighting_t

    elif header.version == header.filelen:
        # This is obviously a mistake in the bsp header, let's assume v1
        # (seen in css map once)
        return dleafambientlighting_t

    else:
        raise LumpVersionUnsupportedError(header.version)



"""
Lump 21 - Area Portals
======================

This lump contains an array of :any:`dareaportal_t`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402

dareaportal_t = Struct(
    'portalKey' / Int16ul,
    'otherArea' / Int16ul,
    'firstClipPortalVert' / Int16ul,
    'clipPortalVerts' / Int16ul,
    'planeNum' / Int32sl
)


@lump_array
@lump_version(0)
def lump_21(header, profile=None):
    return dareaportal_t



"""
Lump 6 - Texture Info
=====================

This lump contains an array of :any:`texinfo_t`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402
from valvebsp.structs.flags import texinfo_flags32  # NOQA: #402

texinfo_t = Struct(
    'textureVecsTexelsPerWorldUnits' / Float32l[2][4],
    'lightmapVecsLuxelsPerWorldUnits' / Float32l[2][4],
    'flags' / texinfo_flags32,
    'texData' / Int32sl * 'index into :ref:`lump 2<lump_2>`',
)


@lump_array
@lump_version(0)
def lump_6(header, profile=None):
    return texinfo_t



"""
Lump 60 - Overlay Fades
=======================

This lump contains an array of :any:`doverlayfade_t`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA #402


doverlayfade_t = Struct(
    'fadeDistMinSq' / Float32l,
    'fadeDistMaxSq' / Float32l
)


@lump_array
@lump_version(0)
def lump_60(header, profile=None):
    return doverlayfade_t



"""
Lump tlpd - Prop Detail Lighting
================================

This lump contains an array of :any:`DetailPropLightStyle_t`.
In bsp, this array is prefixed by an integer representing its length.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402

DetailPropLightStyle_t = Struct(
    'lighting' / ColorRGBExp32,
    'style' / Int8ul,
)


@lump_version(0)
@lump_prefixed_array(Int32sl)
def lump_tlpd(header, profile=None):
    return DetailPropLightStyle_t



"""
Lump 50 - Water Overlays
========================

This lump contains an array of dwateroverlay_t, identical to :any:`doverlay_t`. (:ref:`as seen in lump_45<lump_45>`).
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from valvebsp.structs.common import *  # NOQA #402
from valvebsp.structs.lump_45 import doverlay_t as dwateroverlay_t  # NOQA #402


@lump_array
@lump_version(0)
def lump_50(header, profile=None):
    return dwateroverlay_t



"""
Lump 37 - Primitives
====================

This lump contains an array of :any:`dprimitive_t`.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA #402

dprimitive_type = Enum(
    Int8ul,
    PRIM_TRILIST=0,
    PRIM_TRISTRIP=1,
)

dprimitive_t = Aligned(2, Struct(
    'type' / dprimitive_type,
    'firstIndex' / Int16ul * "index into :ref:`lump 39<lump_39>`",
    'indexCount' / Int16ul,
    'firstVert' / Int16ul * "index into :ref:`lump 38<lump_38>`",
    'vertCount' / Int16ul,
))


@lump_array
@lump_version(0)
def lump_37(header, profile=None):
    return dprimitive_t



"""
Lump 3 - Vertexes
=================

This lump contains an array of :any:`Vector`, three floats corresponding to X/Y/Z coordinates. Vertexes are typically combined to form edges (:ref:`see lump_12<lump_12>`).
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from construct import *  # NOQA: #402
from valvebsp.structs.common import *  # NOQA: #402


@lump_array
@lump_version(0)
def lump_3(header, profile=None):
    return Vector
